import system.io
import .parser.core
import .imp.parser
import .imp.has_to_string
import .imp.repr
import .imp.pretty_python
import .sugar_imp.parser
import .sugar_imp.has_to_string
import .sugar_imp.repr
import .sugar_imp.pretty_python
import .desugaring
import .to_python

def usage : string :=
string.join $ list.intersperse "\n" $ [
    "Usage: Desugar [options] <program>                                   ",
    "                                                                     ",
    "Translation to Python and pretty printing for SugarImp and Imp code. ",
    "                                                                     ",
    "OPTIONS:                                                             ",
    "  --translate translate SugarImp to Imp program                      ",
    "  --Spretty   pretty print SugarImp program                          ",
    "  --Ipretty   pretty print Imp program                               ",
    "  --Spython   translate SugarImp program to Python                   ",
    "  --Ipython   translate Imp program to Python                        ",
    "  --Slean     print SugarImp program in Lean syntax                  ",
    "  --Ilean     print Imp program in Lean syntax                       ",
    "                                                                     "]

@[derive decidable_eq]
inductive flag
| translate
| spretty
| ipretty
| spython
| ipython
| slean
| ilean
| repl
| program : string → flag

instance flag.inhabitated : inhabited flag :=
⟨ flag.translate ⟩

def to_flag : string → flag
| "--translate" := flag.translate
| "--Spretty" := flag.spretty
| "--Ipretty" := flag.ipretty
| "--Spython" := flag.spython
| "--Ipython" := flag.ipython
| "--Slean" := flag.slean
| "--Ilean" := flag.ilean
| "--repl"  := flag.repl
| n := flag.program n

def to_flags : list string → list flag :=
list.map to_flag

open flag

def flag.to_string : flag → string
| translate := "translate"
| spretty := "spretty"
| ipretty := "ipretty"
| spython := "spython"
| ipython := "ipython"
| slean := "slean"
| ilean := "ilean"
| repl := "repl"
| (program n) := "program " ++ to_string n

instance flag.has_to_string : has_to_string flag :=
⟨ flag.to_string ⟩

variable [io.interface]

def parse_file {t : Type} (file_name : string) (p : parser char t) : io t :=
do buf ← io.fs.read_file file_name,
   match p.run buf with
   | sum.inl msg := io.fail msg
   | sum.inr v := return v
   end

def split_flags (fs : list flag) : (list string × list flag) :=
let ps := fs.filter_map (λ p, match p with | flag.program n := some n | _ := none end),
    os := fs.filter_map (λ p, match p with | flag.program _ := none | d := some d end) in
    (ps, os)

def for_each_file : list string → (io.handle → io unit) → io unit
| [] each := return ()
| (f :: fs) each :=
  do h ← io.mk_file_handle f io.mode.read,
     each h,
     for_each_file fs each

def is_repl : list flag → bool
| [] := false
| (f :: fs) := decidable.to_bool (f = flag.repl)

def repl : io unit :=
  io.forever $ do
    io.print "> ",
    ln ← io.get_line,
    let res := sugar.parser.parse_string ln (sugar.parser.fn <* parser.sat (= lexer.tok.eof)) in
    match res with
    | (sum.inl err) := io.print  $ "parse error: " ++ err
    | (sum.inr v) := do
        io.print_ln "----------- Sugar -------------",
        io.print_ln "----------- Lean  -------------",
        io.print_ln (repr v),
        io.print_ln "----------- Pretty ------------",
        io.print_ln (to_string v),
        io.print_ln "----------- Python ------------",
        io.print_ln (has_to_python.to_python v),
        let dv := desugar_func v,
        io.print_ln "----------- Imp   -------------",
        io.print_ln "----------- Lean  -------------",
        io.print_ln (repr dv),
        io.print_ln "----------- Pretty ------------",
        io.print_ln (to_string dv),
        io.print_ln "----------- Python ------------",
        io.print_ln (has_to_python.to_python dv)
    end

def run_or_fail {a : Type} (p : parser char a) (out : char_buffer) : io a :=
match p.run out with
| sum.inl err := io.fail err
| sum.inr v := return v
end

def main : io unit :=
do args ← io.cmdline_args,
   let fs := to_flags args,
   match fs with
   | [] := io.print usage
   | _ :=
     do let (files, flags) := split_flags fs,
        -- io.print_ln $ "Files: " ++ to_string files,
        -- io.print_ln $ "Flags: " ++ to_string flags,
        if is_repl flags
        then repl
        else for_each_file files $ λ h,
          do out ← io.fs.read_to_end h,
             match list.head flags with
             | flag.translate :=
                do p ← run_or_fail sugar.parser.program out,
                   io.print_ln (to_string (desugar_prog p))
             | flag.spretty :=
                do p ← run_or_fail sugar.parser.program out,
                   io.print_ln (to_string p) -- fix me
             | flag.ipretty :=
                do p ← run_or_fail imp.parser.program out,
                   io.print_ln (to_string p)
             | flag.spython :=
                do p ← run_or_fail sugar.parser.program out,
                   io.print_ln (to_py_program p)
             | flag.ipython :=
                do p ← run_or_fail imp.parser.program out,
                   io.print_ln (to_py_program p)
             | flag.slean :=
                do p ← run_or_fail sugar.parser.program out,
                   io.print_ln (repr p)
             | flag.ilean :=
                do p ← run_or_fail imp.parser.program out,
                   io.print_ln (repr p)
             | flag.repl := repl
             | _ := io.fail "unexpected flag"
             end
  end
